/*
  plcLib Version 2.0 Beta 6, last updated 22nd April, 2023.
  
  Simple PLC-style programming for the Arduino and compatibles.

  Authors: 
  
      Walter Ditch - Version 0.5 to present
      Stevan Tosic - Version 2.0 to present

  Licence: MIT
  
  Copyright (c) 2023, the plcLib authors

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

// Uncomment the following line(s) to enable negative logic inputs and/or outputs
// #define negLogicInputs
// #define negLogicOutputs

#include "Arduino.h"
#include "plcLib.h"

extern unsigned long scanValue = 0;

// Define default pin directions and initial output levels.
// (or leave these unconfigured if noPinDefs is set in the user sketch)

void setupPLC() {
    #ifndef noPinDefs
	// Basic input pins
	pinMode(X0, INPUT);
	pinMode(X1, INPUT);
	pinMode(X2, INPUT);
	pinMode(X3, INPUT);
	
	// Basic output pins
	pinMode(Y0, OUTPUT);
	pinMode(Y1, OUTPUT);
	pinMode(Y2, OUTPUT);
	pinMode(Y3, OUTPUT);
    
    // Default output port values
	digitalWrite(Y0, LOW);
	digitalWrite(Y1, LOW);
	digitalWrite(Y2, LOW);
	digitalWrite(Y3, LOW);
    #endif /* noPinDefs */
}

// Digital Read helper function, suppporting negative logic inputs
unsigned long digitalRead2(int input) {

unsigned long value;

#ifdef negLogicInputs 
digitalRead(input) == 1 ? value = 0 : value = 1; // -ve logic
#else
digitalRead(input) == 1 ? value = 1 : value = 0; // +ve logic
#endif

return value;
}

// Digital Write helper function, suppporting negative logic outputs
void digitalWrite2(int output, unsigned int value) {
#ifdef negLogicOutputs
value == 0 ? digitalWrite(output, 1) : digitalWrite(output, 0); // -ve logic
#else
value == 0 ? digitalWrite(output, 0) : digitalWrite(output, 1); // +ve logic
#endif
return;
}

// Read an input pin (pin number supplied as integer)
unsigned long din(int input) {
	scanValue = digitalRead2(input);
	return(scanValue);
}

// Read an auxiliary input (variable supplied as unsigned long)
unsigned long din(unsigned long input) {
	scanValue = input & 1;
	return(scanValue);
}

// Read an inverted input (pin number supplied as integer)
unsigned long dinNot(int input) {
	if (digitalRead2(input) == 1) {
		scanValue = 0;
	}
	else {
		scanValue = 1;
	}
	return(scanValue);
}

// Read an inverted auxiliary input (variable supplied as unsigned long)
unsigned long dinNot(unsigned long input) {
	if (input & 1 == 1) {
		scanValue = 0;
	}
	else {
		scanValue = 1;
	}
	return(scanValue);
}

// Read an analogue input (input pin supplied as an integer)
unsigned long ain(int input) {
	scanValue = analogRead(input);
	return(scanValue);
}

// Read an analogue auxiliary value (variable supplied as an unsigned long)
unsigned long ain(unsigned long input) {
	scanValue = input;
	return(scanValue);
}

// Output to an output pin
unsigned long dout(int output) {
	if (scanValue & 1 == 1) {
		digitalWrite2(output, HIGH);
	}
	else {
		digitalWrite2(output, LOW);
	}
	return(scanValue);
}

// Output to an auxiliary variable (variable type = unsigned long)
unsigned long dout(unsigned long &output) {
	if (scanValue & 1 == 1) {
		output = 1;
	}
	else {
		output = 0;
	}
	return(scanValue);
}

// Output to an output pin (inverted)
unsigned long doutNot(int output) {
	if (scanValue & 1 == 1) {
		digitalWrite2(output, LOW);
	}
	else {
		digitalWrite2(output, HIGH);
	}
	return(scanValue);
}

// Output to an auxiliary variable (inverted) (variable type = unsigned long)
unsigned long doutNot(unsigned long &output) {
	if (scanValue & 1 == 1) {
		output = 0;
	}
	else {
		output = 1;
	}
	return(scanValue);
}

// Output a PWM value to an output pin (scanValue in range 0-1023)
unsigned long pout(int output) {
	analogWrite(output, scanValue / 4);
	return(scanValue);
}

// AND scanValue with input (pin number supplied as integer)
unsigned long andBit(int input) {
	scanValue = scanValue & 1 & digitalRead2(input);
	return(scanValue);
}

// AND scanValue with auxiliary variable (variable supplied as unsigned long)
unsigned long andBit(unsigned long input) {
	scanValue = scanValue & 1 & input;
	return(scanValue);
}

// AND scanValue with inverted input (pin number supplied as integer)
unsigned long andNotBit(int input) {
	scanValue = scanValue & 1 & ~digitalRead2(input);
	return(scanValue);
}

// AND scanValue with inverted auxiliary variable (variable supplied as unsigned long)
unsigned long andNotBit(unsigned long input) {
	scanValue = scanValue & 1 & ~input;
	return(scanValue);
}

// OR scanValue with input (pin number supplied as integer)
unsigned long orBit(int input) {
	scanValue = (scanValue & 1) | digitalRead2(input);
	return(scanValue);
}

// OR scanValue with auxiliary variable (variable supplied as unsigned long)
unsigned long orBit(unsigned long input) {
	scanValue = (scanValue & 1) | input;
	return(scanValue);
}

// OR scanValue with inverted input (pin number supplied as integer)
unsigned long orNotBit(int input) {
	if (scanValue == 1) {
	}
	else {
		if (digitalRead2(input) == 0) {
			scanValue = 1;
		}
		else {
			scanValue = 0;
		}
	}
	return(scanValue);
}

// OR scanValue with inverted auxiliary variable (variable supplied as unsigned long)
unsigned long orNotBit(unsigned long input) {
	if (scanValue == 1) {
	}
	else {
		if (input == 0) {
			scanValue = 1;
		}
		else {
			scanValue = 0;
		}
	}
	return(scanValue);
}

// XOR scanValue with input (pin number supplied as integer)
unsigned long xorBit(int input) {
	scanValue = (scanValue & 1) ^ digitalRead2(input);
	return(scanValue);
}

// XOR scanValue with auxiliary variable (variable supplied as unsigned long)
unsigned long xorBit(unsigned long input) {
	scanValue = (scanValue & 1) ^ input;
	return(scanValue);
}

// XOR scanValue with inverted input (pin number supplied as integer)
unsigned long xorNotBit(int input) {
	scanValue = (scanValue & 1) ^ ~digitalRead2(input);
	return(scanValue);
}

// XOR scanValue with inverted auxiliary variable (variable supplied as unsigned long)
unsigned long xorNotBit(unsigned long input) {
	scanValue = (scanValue & 1) ^ ~input;
	return(scanValue);
}


// Set - Reset latch (output and reset pin numbers supplied as integers) - tested +ve/-ve logic
unsigned long latch(int output, int reset) {
#ifdef negLogicOutputs
	scanValue = (scanValue | ~digitalRead(output)) & 1;	 // Self latch by ORing with Output pin (Q)
#else
	scanValue = scanValue | digitalRead(output);		 // Self latch by ORing with Output pin (Q)
#endif
	scanValue = scanValue & ~digitalRead2(reset);		 // AND with inverted Reset Pin
	if (scanValue == 1) {
		digitalWrite2(output, HIGH);
	}
	else {
		digitalWrite2(output, LOW);
	}
	return(scanValue);
}


// Set - Reset latch (output pin number supplied as integer, reset as unsigned long variable) - tested +ve/-ve logic
unsigned long latch(int output, unsigned long reset) {
#ifdef negLogicOutputs
	scanValue = (scanValue | ~digitalRead(output)) & 1;	 // Self latch by ORing with Output pin (Q)
#else
	scanValue = scanValue | digitalRead(output);		 // Self latch by ORing with Output pin (Q)
#endif
	scanValue = scanValue & ~reset;						 // AND with inverted Reset variable
	if (scanValue == 1) {
		digitalWrite2(output, HIGH);
	}
	else {
		digitalWrite2(output, LOW);
	}
	return(scanValue);
}

// Set - Reset latch (output as unsigned long variable and reset pin as integer)
unsigned long latch(unsigned long &output, int reset) {
	scanValue = scanValue | output;						// Self latch by ORing with Output pin (Q)
	scanValue = scanValue & ~digitalRead2(reset);		// AND-Not with Reset Pin
	if (scanValue == 1) {
		output = 1;
	}
	else {
		output = 0;
	}
	return(scanValue);
}

// Set - Reset latch (output and reset values are unsigned long auxiliary variables)
unsigned long latch(unsigned long &output, unsigned long reset) {
	scanValue = scanValue | output;				// Self latch by ORing with Output variable (Q)
	scanValue = scanValue & ~reset;				// AND-Not with Reset variable
	if (scanValue == 1) {
		output = 1;
	}
	else {
		output = 0;
	}
	return(scanValue);
}
// On delay timer - requires continuously enabled input from previous scanValue
// (timerState and timerPeriod are unsigned long values - 32 bit)
unsigned long timerOn(unsigned long &timerState, unsigned long timerPeriod) {
	if (scanValue == 0) {									// timer is disabled
		timerState = 0;										// Clear timerState (0 = 'not started')
	}
	else {													// Timer is enabled
		if (timerState == 0) {								// Timer hasn't started counting yet
			timerState = millis();							// Set timerState to current time in milliseconds
			scanValue = 0;									// Result = 'not finished' (0)
		}
		else {												// Timer is active and counting
			if (millis() - timerState >= timerPeriod) {		// Timer has finished
				scanValue = 1;								// Result = 'finished' (1)
			}
			else {											// Timer has not finished
				scanValue = 0;								// Result = 'not finished' (0)
			}
		}
	}
	return(scanValue);										// Return result (1 = 'finished',
                                                            // 0 = 'not started' / 'not finished')
}

// Fixed width pulse - enabled by momentary input from previous scanValue 
// (timerState and timerPeriod are unsigned long values - 32 bit)
unsigned long timerPulse(unsigned long &timerState, unsigned long timerPeriod) {
	if (scanValue == 1 || timerState != 0){					// Timer is enabled
		if (timerState == 0) {								// Timer hasn't started counting yet
			timerState = millis();							// Set timerState to current time in milliseconds
			scanValue = 1;									// Pulse = 'Active' (1)
		}
		else {												// Timer is active and counting
			if (millis() - timerState >= timerPeriod) {		// Timer has finished
			    if (scanValue == 0) {                       // Finished AND trigger is low
				    timerState = 0;							// Re-enabled timer
					scanValue = 0;							// Pulse = 'finished' (0)
				}
				else {                                      // Finished but trigger is still high
					scanValue = 0;                          // Wait for trigger to go low before re-enabling
				}
			}
			else {											// Timer has not finished
				scanValue = 1;								// Pulse = 'Active' (1)
			}
		}
	}
	return(scanValue);										// Return result (1 = 'active',
															// 0 = 'not started' / 'not yet re-enabled')
}
// Off delay timer - turns on immediately when enabled, then delays turning off when previous scanValue goes low
// (timerState and timerPeriod are unsigned long values - 32 bit)
unsigned long timerOff(unsigned long &timerState, unsigned long timerPeriod) {
	if (scanValue == 0) {									// Timer input is off (scanValue = 0)
		if (timerState == 0) {								// Timer is not started so do nothing
		}
		else {												// Timer is active and counting
			if (millis() - timerState >= timerPeriod) {		// Timer has finished
				scanValue = 0;								// Result = 'turn-off delay finished' (0)
			}
			else {											// Timer has not finished
				scanValue = 1;								// Result = 'turn-off delay not finished' (1)
			}
		}
	}
	else {													// Timer input is high (scanValue = 1)
		timerState = millis();								// Set timerState to current time in milliseconds
	}
	return(scanValue);										// Return result (1 = 'pulse On' / 'turn-off delay in progress',
															// 0 = 'not started' / 'finished')
}

// Cycle timer - creates a repeating pulse waveform when enabled by previous scanValue
// (timer1Period and timer2Period are unsigned long values - 32 bit)
unsigned long timerCycle(Waveform &timerName, unsigned long timer1Period, unsigned long timer2Period) {
	if (scanValue == 0) {									// Enable input is off (scanValue = 0)
		timerName.elapsedHigh = 0;							// Ready to start LOW pulse period when enabled
		timerName.elapsedLow = 1;
	}
	else {													// Enabled
		if (timerName.elapsedHigh == 0) {					// Low pulse Active
			if (timerName.elapsedLow == 1) {				// LOW pulse period starting
				timerName.elapsedLow = millis();			// Set timerState to current time in milliseconds
			}
			else if (millis() - timerName.elapsedLow >= timer1Period) {	// Low pulse period has finished
				timerName.elapsedLow = 0;	
				timerName.elapsedHigh = 1;					// Ready to start HIGH pulse period
			}
			scanValue = 0;									// Result = 'Pulse LOW' (0)
		}
		if (timerName.elapsedLow == 0) {					// High pulse Active
			if (timerName.elapsedHigh == 1) {				// HIGH pulse period starting
				timerName.elapsedHigh = millis();			// Set timerState to current time in milliseconds
			}
			else if (millis() - timerName.elapsedHigh >= timer2Period) {	// High pulse has finished
				timerName.elapsedHigh = 0;	
				timerName.elapsedLow = 1;					// Ready to start LOW pulse period
			}
			scanValue = 1;									// Result = 'Pulse HIGH' (1)
		}
	}
	return(scanValue);
}

// Test whether an analogue input is greater than a second analogue input
unsigned long compareGT(int input) {
	if (scanValue > analogRead(input)) {
		scanValue = 1;
	}
	else {
		scanValue = 0;
	}
	return(scanValue);
}

// Test whether an analogue input is greater than a fixed unsigned long value
unsigned long compareGT(unsigned long input) {
	if (scanValue > input) {
		scanValue = 1;
	}
	else {
		scanValue = 0;
	}
	return(scanValue);
}

// Test whether an analogue input is less than a second analogue input
unsigned long compareLT(int input) {
	if (scanValue < analogRead(input)) {
		scanValue = 1;
	}
	else {
		scanValue = 0;
	}
	return(scanValue);
}

// Test whether an analogue input is less than a fixed unsigned long value
unsigned long compareLT(unsigned long input) {
	if (scanValue < input) {
		scanValue = 1;
	}
	else {
		scanValue = 0;
	}
	return(scanValue);
}

// Set a latched output (output pin number supplied as integer)
unsigned long setL(int output) {
	if (scanValue == 1) {
		digitalWrite2(output, HIGH);
	}
#ifdef negLogicOutputs 
    digitalRead(output) == 1 ? scanValue = 0 : scanValue = 1; // -ve logic
#else
    digitalRead(output) == 1 ? scanValue = 1 : scanValue = 0; // +ve logic
#endif
	return(scanValue);
}

// Set a latched output (variable supplied as unsigned long)
unsigned long setL(unsigned long &output) {
	scanValue = scanValue | output;		// Self latch by ORing with Output variable
	if (scanValue == 1) {
		output = 1;
	}
	return(scanValue);
}

// reset (or clear) a latched output (output pin number supplied as integer)
unsigned long resetL(int output) {
	if (scanValue == 1) {
		digitalWrite2(output, LOW);
	}
#ifdef negLogicOutputs 
    digitalRead(output) == 1 ? scanValue = 0 : scanValue = 1; // -ve logic
#else
    digitalRead(output) == 1 ? scanValue = 1 : scanValue = 0; // +ve logic
#endif
	return(scanValue);
}

// reset (or clear) a latched output (variable supplied as unsigned long)
unsigned long resetL(unsigned long &output) {
	if (scanValue == 1) {
		output = 0;
	}
	return(scanValue);
}

// Up, down, or up-down counter
Counter::Counter(unsigned long pv)	// Counter constructor method
{									// (Default values are for an up counter)
	_pv = pv;						// Set preset value using supplied parameter
	_val = 0;						// Running count = zero
	_uQ = 0;						// Up counter upper Q output = 0
	_lQ = 1;						// Down counter lower Q output = 1
	_ctUpEdge = 0;					// Prepare rising edge detect for up counter
	_ctDownEdge = 0;				// Prepare rising edge detect for down counter
}

Counter::Counter(unsigned long pv, unsigned long direction)	// Counter constructor method
{									// (Second parameter sets default direction)
	_pv = pv;						// Set preset value using supplied parameter
	if(direction == 0){				// Set start values for an up counter
		_val = 0;					// Running count = zero
		_uQ = 0;					// Up counter upper Q output = 0
		_lQ = 1;					// Down counter lower Q output = 1
	}
	else {							// Set start values for a down counter
		_val = _pv;					// Running count = preset value
		_uQ = 1;					// Up counter upper Q output = 1
		_lQ = 0;					// Down counter lower Q output = 0
	}
	_ctUpEdge = 0;					// Prepare rising edge detect for up counter
	_ctDownEdge = 0;				// Prepare rising edge detect for down counter
}

unsigned long Counter::value()
{
   scanValue = _val;
   return _val;
}

unsigned long Counter::setValue(unsigned long val)
{
    _val = val;
    return _val;
}
	
void Counter::clear()				// Clear counter method
{
	if(scanValue == 1) {			// Enabled if scanValue = 1
		_val = 0;					// Running count = 0
		_uQ = 0;					// Up counter upper Q output = 0
		_lQ = 1;					// Down counter lower Q output = 1
		_ctUpEdge = 0;				// Prepare rising edge detect for up counter
		_ctDownEdge = 0;			// Prepare rising edge detect for down counter
	}
}

void Counter::preset()				// Preset counter method
{
	if(scanValue == 1) {			// Enabled if scanValue = 1
		_val = _pv;					// Running count = preset value
		_uQ = 1;					// Up counter upper Q output = 1
		_lQ = 0;					// Down counter lower Q output = 0
		_ctUpEdge = 0;				// Prepare rising edge detect for up counter
		_ctDownEdge = 0;			// Prepare rising edge detect for down counter
	}
}

unsigned long Counter::upperQ()		// Read up counter upper Q output method
{
	if (_uQ == 1){					// Set scanValue = 1 if upper Q = 1
		scanValue = 1;
	}
	else {
		scanValue = 0;				// Otherwise set scanValue = 0
	}
	return(_uQ);					// Return upper Q value
}

unsigned long Counter::lowerQ()		// Read down counter lower Q output method
{
	if (_lQ == 1){					// Set scanValue = 1 if lower Q = 1
		scanValue = 1;
	}
	else {
		scanValue = 0;				// Otherwise set scanValue = 0
	}
	return(_lQ);					// Return lower Q value
}

void Counter::countUp()				// Count up method
{
	if (_val != _pv) {      			// Not yet finished counting so continue
		if (scanValue == 0) {		// clock = 0 so clear counter edge detect
			_ctUpEdge = 0;
		}
		else {						// Clock = 1
			if (_ctUpEdge == 0) {	// Rising edge detected so increment count
				_ctUpEdge = 1;		// Set counter edge
				_val++;				// Increment count
				if(_val == _pv) {	// Counter has reached final value
					_uQ = 1;		// Counter upper Q output = 1
					_lQ = 0;		// Counter lower Q output = 0
				}
				if(_val != _pv){	// Counter is not yet finished
					_uQ = 0;		// Counter upper Q output = 0
					_lQ = 0;		// Counter lower Q output = 0
				}
			}
		}
	}
}

void Counter::countDown()			// Count down method
{
	if (_val != 0) {					// Not yet finished so continue
		if (scanValue == 0) {		// clock = 0 so clear counter edge detect
			_ctDownEdge = 0;
		}
		else {						// Clock = 1
			if (_ctDownEdge == 0) {	// Rising edge detected so decrement count
				_ctDownEdge = 1;	// Set counter edge
				_val--; 			// Decrement count
				if(_val == 0) {		// Counter has reached final value
					_uQ = 0;		// Counter QUp output = 0
					_lQ = 1;		// Counter QDown output = 1
				}
				if(_val != 0) {		// Counter is not yet finished
					_uQ = 0;		// Counter upper Q output = 0
					_lQ = 0;		// Counter lower Q output = 0
				}
			}
		}
	}
}

// Shift register
Shift::Shift()						// Shift register constructor method
{									// (Register width = 32 bits)
	_val = 0;						// Set the shift register to zero
	_srLeftEdge = 0;				// Prepare rising edge detect for left shift
	_srRightEdge = 0;				// Prepare rising edge detect for right shift
}

Shift::Shift(unsigned long val)	    // Shift register constructor method
{									// (Register width = 32 bits)
	_val = val;					    // Set initial value
	_srLeftEdge = 0;				// Prepare rising edge detect for left shift
	_srRightEdge = 0;				// Prepare rising edge detect for right shift
}

unsigned long Shift::bitValue(unsigned int bitno)	// Read a bit at a specified position
{
	if(bitRead(_val, bitno) == 1) {
		scanValue = 1;				// Tested bit = 1
	}
	else {
		scanValue = 0;				// Tested bit = 0
	}
	return(scanValue);				// Return tested bit value
}

unsigned long Shift::value()		// Return the current shift register value
{
	scanValue = _val;
    return(_val);
}

unsigned long Shift::setValue(unsigned long val)
{
    _val = val;
    return _val;
}

void Shift::reset()					// Reset the shift register if scanValue = 0
{
	if(scanValue == 1){
		_val = 0;					// Set the shift register to zero
		_srLeftEdge = 0;			// Prepare rising edge detect for left shift
		_srRightEdge = 0;			// Prepare rising edge detect for right shift
		}
}

void Shift::inputBit()				// Set the input bit of the shift register
{
	if (scanValue == 0) {			// If scanValue = 0, clear input bit
		_inbit = 0;
	}
	else {							// Otherwise set the input bit
		_inbit = 1;
	}
}

void Shift::shiftRight()			// Shift right method
{
	if (scanValue == 0) {			// clock = 0 so clear shift right edge detect
		_srRightEdge = 0;
	}
	else {							// Clock = 1
		if (_srRightEdge == 0) {	// Rising edge detected so shift right
			_srRightEdge = 1;		// Set shift right edge detect
			_val = _val >> 1;		// Shift to the right
			if (_inbit == 1) {		// Shift-in new input bit at LHS
				bitSet(_val, 15);
			}
		}
	}
}

void Shift::shiftLeft()				// Shift left method
{
	if (scanValue == 0) {			// clock = 0 so clear shift left edge detect
		_srLeftEdge = 0;
	}
	else {							// Clock = 1
		if (_srLeftEdge == 0) {		// Rising edge detected so shift left
			_srLeftEdge = 1;		// Set shift left edge detect
			_val = _val << 1;		// Shift to the left
			if (_inbit == 1) {		// Shift-in new input bit at RHS
				bitSet(_val, 0);
			}
		}
	}
}

// Single-bit Software Stack
Stack::Stack()						// Stack constructor method
{									// (Register width = 32 bits)
	_val = 0;						// Set the stack to zero
}

void Stack::push()					// Push scanValue bit onto the stack method
{
	_val = _val << 1;				// Shift stack 1-bit to the left
	if (scanValue == 1) {			// Shift-in scanValue bit at RHS
		bitSet(_val, 0);
	}
	else {
		bitClear(_val, 0);
	}
}

void Stack::pop()					// Pop scanValue bit from the stack method
{
	scanValue = bitRead(_val, 0);	// Set scanValue to rightmost bit of stack
	_val = _val >> 1;				// Shift stack 1-bit to the right
}

void Stack::orBlock()				// OR scanValue with value Popped from stack method
{
	scanValue = scanValue | bitRead(_val, 0);	// OR scanValue with top of stack
	_val = _val >> 1;				// Shift stack 1-bit to the right
}

void Stack::andBlock()				// AND scanValue with value Popped from stack method
{
	scanValue = scanValue & bitRead(_val, 0);	// AND scanValue with top of stack
	_val = _val >> 1;				// Shift stack 1-bit to the right
}

unsigned long Stack::value()	    // Return the current Stack value (for debugging only)
{
	scanValue = _val;
    return(_val);
}

// Stack::setValue method is not implemented

// Single scan cycle Pulse with rising or falling edge detection
Pulse::Pulse()						// Pulse constructor method
{
	_val = 0;				        // Set pulse input tracker to zero
	_pulseUpEdge = 0;				// Prepare rising edge detect
	_pulseDownEdge = 0;				// Prepare falling edge detect
    _polarity = 0;                  // Rising edge is default polarity
}

Pulse::Pulse(unsigned int polarity)	// Pulse constructor method
{
	_val = 0;				        // Set pulse input tracker to zero
	_pulseUpEdge = 0;				// Prepare rising edge detect
	_pulseDownEdge = 0;				// Prepare falling edge detect
    if(polarity == 0) {
        _polarity = 0;              // Set rising edge as default polarity
    }
    else {
        _polarity = 1;              // Set falling edge as default polarity
    }
}

void Pulse::inClock()				// Read the clock input method
{
	if (scanValue != _val) {	    // Rising or falling edge detected
		if (scanValue == 1) {		// Rising edge detected
			_pulseUpEdge = 1;		// Set rising edge detect value
			_pulseDownEdge = 0;		// Clear falling edge detect value
			_val = 1;		        // Pulse input tracker = 1
		}
		else {						// Falling edge detected
			_pulseUpEdge = 0;		// Clear rising edge detect value
			_pulseDownEdge = 1;		// Set falling edge detect value
			_val = 0;		        // Pulse input tracker = 0
		}
	}
	else {							// No change detected
	_pulseUpEdge = 0;				// Set both edge detect values to zero
	_pulseDownEdge = 0;				// (and leave pulse tracker unchanged)
	}
}

void Pulse::rising()		        // Pulse rising edge detected method
{
    scanValue = _pulseUpEdge;	    // scanValue = 1 if rising edge detected, 0 otherwise
}

void Pulse::falling()		        // Pulse falling edge detected method
{
	scanValue = _pulseDownEdge;	    // scanValue = 1 if falling edge detected, 0 otherwise
}

unsigned long Pulse::value()	    // Return the current Pulse edge value
{
    if(_polarity == 0) {            // Rising edge is selected default polarity
        scanValue = _pulseUpEdge;
        return _pulseUpEdge;
    }
    else {
        scanValue = _pulseDownEdge; // Falling edge is selected default polarity
        return _pulseDownEdge;
    }
}

// Pulse::setValue and Pulse::value methods are not implemented

void Pulse::expired()
{
    if (scanValue == 1) {
        _pulseUpEdge = 0;           // Set both edge detect values to zero
        _pulseDownEdge = 0;         // (and leave pulse tracker unchanged)
        scanValue = 0;              // Disable subsequent commands in branch
    }
}

State::State() {

        eEdge = 0;           // Entering edge (1 = active)
        lEdge = 0;           // Leaving edge (1 = active)
        _val = 0;            // Enabled (1) / disabled (0)
        history_ptr = NULL;  // NULL = no previous state
    }

State::State(unsigned long val) {

        eEdge = 0;           // Entering edge (1 = active)
        lEdge = 0;           // Leaving edge (1 = active)
        history_ptr = NULL;  // NULL = no previous state
      
        if (val) {        // Value (1 = state active, 0 = inactive)
            eEdge = 1;
            lEdge = 0;
            _val = 1;
        }
        else {
            eEdge = 0;
            lEdge = 0;
            _val = 0;
        }
    }

unsigned long State::value() {
        scanValue = _val;
        return _val;
}

unsigned long State::active() {
        scanValue = _val;
        return _val;
}
    
void State::enable() {                     // Enable the state (1) and set Entering edge (1)
        if (_val == 0) {
            if (scanValue == 1) {
                eEdge = 1;
                lEdge = 0;
                _val = 1;
            }
        }
    }
    
void State::disable() {                    // Disable the state (0) and set Leaving edge (1)
        if (_val == 1) {
            if (scanValue == 1) {
                eEdge = 0;
                lEdge = 1;
                _val = 0;
            }
        }
}
    
unsigned long State::entry() {             // Test if entering (eEdge = 1?), cleared after reading
        if(eEdge == 1) {
            eEdge = 0;
            scanValue = 1;
            return 1;
        }
        else {
            scanValue = 0;
            return 0;
        }
}
    
unsigned long State::exit() {              // Test if leaving (lEdge = 1?), cleared after reading
        if(lEdge == 1) {
            lEdge = 0;
            scanValue = 1;
            return 1;
        }
        else {
            scanValue = 0;
            return 0;
        }
}

void State::back() {                       // Revert back to the previous state, if any
    if (scanValue & 1 == 1) {
        if(_val == 1) {
            if (history_ptr != NULL) {
                eEdge = 0;
                lEdge = 1;
                _val = 0;
                history_ptr->enable();
                history_ptr = NULL;
            }
        }
        else {
            scanValue = 0;
        }
    }
}

// Digital I/O and single bit logical commands based on States

// Read a state value
unsigned long din(State &s) {
	scanValue = s.value() & 1;
	return(scanValue);
}

// Read an inverted state value
unsigned long dinNot(State &s) {
	if (s.value() & 1 == 1) {
		scanValue = 0;
	}
	else {
		scanValue = 1;
	}
	return(scanValue);
}

// Output to a state value
unsigned long dout(State& s) {
	if (scanValue & 1 == 1) {
        s.enable();
	}
	else {
        s.disable();
	}
	return(scanValue);
}

// Output to a state (inverted)
unsigned long doutNot(State& s) {
	if (scanValue == 1) {
        s.disable();
	}
	else {
        s.enable();
	}
	return(scanValue);
}

// AND scanValue with state value
unsigned long andBit(State& s) {
	scanValue = scanValue & s.value();
	return(scanValue);
}

// AND scanValue with inverted state value
unsigned long andNotBit(State& s) {
	scanValue = scanValue & ~s.value();
	return(scanValue);
}

// OR scanValue with state value
unsigned long orBit(State& s) {
	scanValue = scanValue | s.value();
	return(scanValue);
}

// OR scanValue with inverted state value
unsigned long orNotBit(State& s) {
	scanValue = scanValue | ~s.value();
    return(scanValue);                 
}

// XOR scanValue with state value
unsigned long xorBit(State& s) {
	scanValue = scanValue ^ s.value();
	return(scanValue);
}

// XOR scanValue with inverted state value
unsigned long xorNotBit(State& s) {
	scanValue = scanValue ^ ~s.value();
	return(scanValue);
}

// Updated transition command

// Level-based state transition from 'src' to 'dst' (renamed from transL)
unsigned long trans(State& src, State& dst) {
    if (scanValue & 1 == 1) {
        if(src.value() == 1) {
            src.disable();
            dst.enable();
            dst.history_ptr = &src;
            scanValue = 1;
        }
        else {
            scanValue = 0;
        }
	}
	else {
        scanValue = 0;
	}
	return(scanValue);
}

/// OOP
/*
Pin::Pin() // : Object()
{
}

Pin::Pin(int id) // : Object(id)
{
   this->_id = id;
}
*/

// Pin::Pin(int id, int mode, int type = 0) // : Object(id)
Pin::Pin(int id, int mode, int type) // : Object(id)
{
   this->_id = id;
   _mode = mode;
   _type = type;
   pinMode(this->_id, mode);
}

unsigned long Pin::value()
{
   scanValue = digitalRead2(this->_id);
   return scanValue;
}

unsigned long Pin::setValue(unsigned long val)
{
    digitalWrite2(this->_id,  val);
    return val;
}

unsigned long Pin::mode()
{
   return _mode;
}

unsigned long Pin::type()
{
   return _type;
}

Object::Object()	// constructor method with supplied value (default = 0)
{
	_val = 0;						// Set value to 0
}

Object::Object(unsigned long val = 0)	// constructor method with supplied value (default = 0)
{
	_val = val;						// Set value using supplied parameter
}

unsigned long Object::value()    // Read internal value
{
	scanValue = _val;
    return(_val);
}

unsigned long Object::setValue(unsigned long val)
{
    _val = val;
	return(this->_val);
}

// Read an input supplied as Object
unsigned long din(Object &input)
{
	scanValue = input.value() & 1;
	return(scanValue);
}

// Read an inverted input supplied as Object
unsigned long dinNot(Object &input) {

	if (input.value() == 1) {
		scanValue = 0;
	}
	else {
		scanValue = 1;
	}
	return(scanValue);
}

// Output to an output Object
unsigned long dout(Object &output)
{
	if (scanValue & 1 == 1) {
        output.setValue(HIGH);
	}
	else {
        output.setValue(LOW);
	}
	return(scanValue);

}

// Output to an output Object (inverted)
unsigned long doutNot(Object &output) {

	if (scanValue & 1 == 1) {
        output.setValue(LOW);
	}
	else {
        output.setValue(HIGH);
	}

	return(scanValue);
}
